@@ -1,9 +1,10 @@
+from __future__ import annotations
 import re
 
-from .. import h, messages as m
+from .. import h, messages as m, t
 
 
-def accidental2119(doc):
+def accidental2119(doc: t.SpecT) -> None:
     """
     Looks for usage of 2119 keywords in non-normative sections.
     You can override this, allowing the keyword,
@@ -14,11 +15,11 @@ def accidental2119(doc):
         return
     keywords = r"\b(may|must|should|shall|optional|recommended|required)\b"
 
-    def searchFor2119(el):
-        if h.isNormative(el, doc):
+    def searchFor2119(el: t.ElementT) -> None:
+        if h.isNormative(doc, el):
             # 2119 is fine, just look at children
             pass
-        elif h.hasClass(el, "allow-2119"):
+        elif h.hasClass(doc, el, "allow-2119"):
             # Override 2119 detection on this element's text specifically,
             # so you can use the keywords in examples *describing* the keywords.
             pass
